# DO NOT CHANGE THIS!!!
SUN = {
    "system": {
        "inclination": "90 deg",
        "rotation_period": "25.38 d",
        "gamma": "0 m/s",
        "reference_time": 0.0,
        "phase_shift": 0.0,
        "distance": "10 pc"
    },
    "star": {
        "mass": "1.0 solMass",
        "t_eff": "5772 K",
        "metallicity": 0.0,
        "equivalent_radius": "1.0 solRad"
    }
}
