# DO NOT CHANGE THIS!!!
# data from Yoon et al., 2009, AJ, 708, 1
VEGA = {
    "system": {
        "inclination": "4.975 deg",
        "rotation_period": "0.604 d",
        "gamma": "-13000 m/s",
        "reference_time": 0.0,
        "phase_shift": 0.0,
        "distance": "10 pc"
    },
    "star": {
        "mass": "2.135 solMass",
        "t_eff": "10059 K",
        "metallicity": -0.39,
        "equivalent_radius": "2.657 solRad"
    }
}
